# vim: encoding=utf-8 ts=4 et sts=4 sw=4 tw=79 fileformat=unix nu wm=2
"""Helper methods for testing."""
import contextlib
from io import StringIO
from pathlib import Path
from typing import List, Optional, Tuple, Type

import pytest

from a28 import utils
from a28.cli import main


class Helpers:
    """A set of helper utilities used for testing."""

    @staticmethod
    def execute_error(
        command: List[str],
        location: Optional[Path] = None
    ) -> Tuple[str, type, int]:
        """Execute the main a28 function with the provided arguments."""
        temp_stdout = StringIO()

        if location:
            utils.STORAGE = location

        with pytest.raises(SystemExit) as excinfo:
            with contextlib.redirect_stderr(temp_stdout):
                main(command)

        exit_code = excinfo.value.code
        raise_type = excinfo.type

        tmp_output = temp_stdout.getvalue().strip()
        if len(tmp_output.splitlines()):
            output = temp_stdout.getvalue().strip().splitlines()[0]
        else:
            output = tmp_output

        return output, raise_type, exit_code

    @staticmethod
    def execute(
        command: List[str],
        location: Optional[Path] = None
    ) -> Tuple[str, type, int]:
        """Execute the main a28 function with the provided arguments."""
        temp_stdout = StringIO()

        if location:
            utils.STORAGE = location

        with pytest.raises(SystemExit) as excinfo:
            with contextlib.redirect_stdout(temp_stdout):
                main(command)

        exitcode = excinfo.value.code
        raise_type = excinfo.type

        tmp_output = temp_stdout.getvalue().strip()
        if len(tmp_output.splitlines()):
            output = temp_stdout.getvalue().strip().splitlines()[0]
        else:
            output = tmp_output

        return output, raise_type, exitcode


@pytest.fixture()
def helpers() -> Type[Helpers]:
    """Return the helper methods used for testing."""
    return Helpers



"""All the utils functions."""
import os
import re
import sys

# set platform currently being used and major and minor Python version
PLATFORM = sys.platform
PYTHON_MAJOR = sys.version_info.major
PYTHON_MINOR = sys.version_info.minor

# set platform type (use startswith to handle old vers with ver nums)
POSIX = os.name == 'posix'
WINDOWS = os.name == 'nt'
LINUX = sys.platform.startswith('linux')
MACOS = sys.platform.startswith('darwin')
OSX = MACOS  # deprecated alias (thanks David F for the reminder)
FREEBSD = sys.platform.startswith('freebsd')
OPENBSD = sys.platform.startswith('openbsd')
NETBSD = sys.platform.startswith('netbsd')
BSD = FREEBSD or OPENBSD or NETBSD
SUNOS = sys.platform.startswith(('sunos', 'solaris'))
AIX = sys.platform.startswith('aix')

# get the users home path for storing files
if POSIX:
    HOMEDIR = os.path.expanduser('~')
elif WINDOWS:
    HOMEDIR = os.environ['APPDATA']
else:
    # unsupported operating system
    HOMEDIR = os.path.realpath(os.sep)

# set the storage constant
if MACOS:
    APP_SUP = os.path.join(HOMEDIR, 'Library', 'Application Support')
    STORAGE = os.path.join(APP_SUP, 'Area28', 'core')
elif WINDOWS:
    STORAGE = os.path.join(HOMEDIR, 'Area28', 'core')
elif POSIX:
    STORAGE = os.path.join(HOMEDIR, '.config', 'area28', 'core')
else:
    STORAGE = os.path.join(HOMEDIR, '.area28', 'core')

# set the devkit configuration location
CONFIG_PATH = os.path.realpath(os.path.join(STORAGE, os.pardir, 'devkit'))
CONFIG = os.path.join(CONFIG_PATH, 'config.json')


def message(msg='', flush=False, sep=' ', end='\n'):
    """Display a message to the user."""
    print(msg, sep=sep, end=end, flush=flush)


def confirm(msg='confirm?'):
    """Ask the user to confirm their action.

    Ask the user to confirm that ALL files and directories should be
    deleted from the directory specified by the storage variable. The
    STORAGE variable changes depending on the operating system used.
    """
    return bool(
        input(msg) in (
            'y',
            'Y',
            'yes',
            'ok',
            'sure',
            'alrighty',
            'of course',
            'yup', 'oui'
        )
    )


def valid_string(value, min_length=3, max_length=64):
    """Validate string input.

    Check if a string contains only alphanumeric, hyphen, and underscore
    characters and limited to 64 characters.
    """
    check = rf'^[\w_-]{{{min_length},{max_length}}}$'
    return re.match(check, value) is not None



"""Manage account actions.

Using the requests module, authenticate an account on the API and store
the authentication token within a local file.
"""
from argparse import Namespace, _SubParsersAction
from getpass import getpass

from a28 import utils
from a28.api import API, ApiError
from a28.config import ConfigFile


def cli_options(main_parser: _SubParsersAction) -> None:
    """Account subcommand arguments."""
    parser = main_parser.add_parser(
        'account',
        aliases=['acc'],
        help='account actions'
    )
    subparser = parser.add_subparsers(
        dest='account',
        required=True,
        help='account',
    )
    parser_authenticate = subparser.add_parser(
        'authenticate',
        help='authenticate with the endpoint.',
    )
    parser_authenticate.set_defaults(func=authenticate_interactive)
    parser_authenticate.add_argument(
        '-e',
        '--endpoint',
        default=API.DEFAULT_ENDPOINT,
        help='endpoint to use.',
    )
    parser_authenticate.add_argument(
        '-u',
        '--email',
        required=True,
        help='your email address.',
    )
    parser_authenticate.add_argument(
        '-p',
        '--password',
        help='your password.',
    )


def authenticate_interactive(args: Namespace) -> None:
    """Get the user's password."""
    if args.password:
        message = (
            'Warning: Using a password on the command line'
            ' interface can be insecure.'
        )
        utils.message(message)
        password = args.password
    else:
        password = getpass(prompt='Enter your password:')

    authenticate(args.endpoint, args.email, password)


def authenticate(region: str, email: str, password: str) -> None:
    """Authenticate to the API."""
    api = API(region)

    try:
        api.post(url='/account/signin', json_payload={
            'email': email,
            'password': password
        })

        # save the email for reference
        # token will be saved by req hook (see _check_token function)
        with ConfigFile() as data:
            data[api.region]['email'] = email

        utils.message('authentication successful')
    except ApiError as error:
        utils.message(f'error authenticating: {error}')
        raise



# vim: encoding=utf-8 ts=4 et sts=4 sw=4 tw=79 fileformat=unix nu wm=2
"""Test standard command line arguments."""
from a28 import __version__
from tests.conftest import Helpers


def test_a28_no_param(helpers: Helpers) -> None:
    """Test the CLI without any parameters."""
    command = []
    expect = (
        'usage: a28 [-h] [-v]'
        ' {api,package,pkg,system,sys,account,acc} ...'
    )

    output, raise_type, exit_code = helpers.execute_error(command)

    assert raise_type == SystemExit
    assert exit_code == 2
    assert output[0:len(expect)] == expect


def test_a28_version_param(helpers: Helpers) -> None:
    """Check the correct version number is returned."""
    command = ['-v']
    expect = f'a28 version {__version__}'

    output, raise_type, exit_code = helpers.execute(command)

    assert raise_type == SystemExit
    assert exit_code == 0
    assert output[0:len(expect)] == expect


def test_a28_system_exit_on_exception(helpers: Helpers) -> None:
    """Test an exeception being raised.

    Command line call that fail would trigger a system exit with code
    other than 0.
    """
    command = ['account', 'authenticate', '-u', 'joe@example.com', '-p', '234']
    _, raise_type, exit_code = helpers.execute_error(command)

    assert raise_type == SystemExit
    assert exit_code == 1



"""Manage actions relating to the system running area28.

The actions provided by the system module adds the ability to manage the
state of the computer system running Area28 allowing for developers to
maintain the configuration and paths of the system.
"""
import shutil
from argparse import Namespace, _SubParsersAction
from pathlib import Path

from a28 import utils


def cli_options(mainparser: _SubParsersAction) -> None:
    """System subcommand arguments.

    Define the arguments that the system subcommand can accept. The
    system subcommand supports the following commands:

    clean: Clean all files from the current system.
    exists: Return if the system currently contains a configuration.
    path: Return the current system configuration path.
    """
    parser = mainparser.add_parser(
        'system',
        aliases=['sys'],
        help='system actions'
    )
    subparser = parser.add_subparsers(
        dest='system',
        required=True,
        help='system',
    )
    env_group = parser.add_mutually_exclusive_group(required=True)
    env_group.add_argument(
        '--environment', '-e',
        dest='environment',
        required=False,
        default='dev',
        choices=['prod', 'staging', 'dev'],
        help='switch the environment to use'
    )
    env_group.add_argument(
        '--all', '-a',
        required=False,
        action='store_true',
        help='remove all environments'
    )
    parser_exist = subparser.add_parser(
        'exists',
        help='check if the system configuration exists.',
    )
    parser_exist.set_defaults(func=exists)
    parser_clean = subparser.add_parser(
        'clean',
        help='clean (delete) the configuration permanently.',
    )
    parser_clean.set_defaults(func=clean)
    parser_clean.add_argument(
        '-f',
        '--force',
        action='store_true',
        help='force the configuration to be removed bypassing confirmation',
    )
    parser_path = subparser.add_parser(
        'path',
        help='return the location of the configuration path.',
    )
    parser_path.set_defaults(func=path)
    parser_path.add_argument(
        '-m',
        '--minimal',
        action='store_true',
        help='return a minimal version of the path to use in scripts',
    )


def clean_at_path(config_path: str, force_remove: bool = False) -> None:
    """Clean all the config files relating to Area28.

    Using the shutil utilities, delete all the files provided in the
    config_path directory. If an error occurs, catch the exception and
    print it out to the STDOUT.

    Args:
        args (Namespace): The list of provided arguments.

    Returns:
        None: Does not return any value.
    """
    message = f'Delete ALL configuration data from "{config_path}"? y/n?'

    if not force_remove and not utils.confirm(message):
        utils.message(f'Not deleting {config_path}')
        return

    # else the user wants to delete the config at the specified path
    if not Path(config_path).is_dir():
        utils.message(f'No configuration to clean at "{config_path}".')
        raise FileNotFoundError

    try:
        shutil.rmtree(config_path)
        utils.message(f'Deleted all files and directories in {config_path}')
    except OSError as e:
        utils.message(f'Error: {config_path} : {e.strerror}')
        raise


def clean(args: Namespace) -> None:
    """Clean all the configuration files relating to Area28.

    Using the shutil utilities,
    if all is specified, ask and delete all the env [prod, dev, stagin]
    accordingly, or else, remove all the files provided in the args.path
    directory. If an error occurs, catch the exception and print it out
    to the STDOUT.

    Args:
        args (Namespace): The list of provided arguments.

    Returns:
        None: Does not return any value.
    """
    environment = args.environment if not args.all else ''
    environment_path = Path(utils.STORAGE) / environment
    clean_at_path(environment_path, force_remove=args.force)


def exists(args: Namespace) -> None:
    """Check if the configuration directory exists.

    Using the path provided in the arguments, check if the configuration
    path exists. If no path is provided, the default path defined in the
    system utility is used.

    Args:
        args (Namespace): The list of provided arguments.

    Returns:
        None: Does not return any value.
    """
    environment = args.environment if not args.all else ''
    environment_path = Path(utils.STORAGE) / environment
    if environment_path.is_dir():
        utils.message(f'Configuration exists at "{environment_path}".')
    else:
        utils.message(f'No configuration exists at "{environment_path}".')


def path(args: Namespace) -> None:
    """Print the path of the system configuration.

    Print the path currently used to store the system configuration
    files. If the path option is provided in the ArgParse Namespace, the
    provided path will be returned.

    If the minimal ArgParse namespace is provided and set to boolean
    True, the output will be reduced to only the path.

    It does NOT change the configuration path

    Args:
        args (Namespace): The list of provided arguments.

    Returns:
        None: Does not return any value.
    """
    environment = args.environment if not args.all else ''
    environment_path = Path(utils.STORAGE) / environment
    if args.minimal:
        utils.message(f'{environment_path}')
    else:
        utils.message(f'Configuration path set to "{environment_path}".')



# vim: encoding=utf-8 ts=4 et sts=4 sw=4 tw=79 fileformat=unix nu wm=2
"""Run package command tests."""
import argparse

from a28.package import generate_jsondata
from tests.conftest import Helpers


def test_a28_package_no_param(helpers: Helpers) -> None:
    """Ensure all subcommands are present."""
    command = ['package']
    expect = 'usage: a28 package [-h] {init,meta,build,install,publish} ...'

    output, raise_type, exit_code = helpers.execute_error(command)

    assert raise_type == SystemExit
    assert exit_code == 2
    assert output[0:len(expect)] == expect


def test_a28_init_no_param(helpers: Helpers) -> None:
    """Test package initialization without parameters."""
    command = ['package', 'init']
    expect = 'usage: a28 package init [-h] -s SCOPE -n NAME -t'

    output, raise_type, exit_code = helpers.execute_error(command)

    assert raise_type == SystemExit
    assert exit_code == 2
    assert output[0:len(expect)] == expect


def test_a28_build_no_param(helpers: Helpers) -> None:
    """Test package build without parameters."""
    command = ['package', 'build']
    expect = 'usage: a28 package build [-h] --src SRC [--dest DEST]'

    output, raise_type, exit_code = helpers.execute_error(command)

    assert raise_type == SystemExit
    assert exit_code == 2
    assert output[0:len(expect)] == expect


def test_a28_install_no_param(helpers: Helpers) -> None:
    """Test package install without parameters."""
    command = ['package', 'install']
    expect = 'usage: a28 package install [-h] --pkg PKG'

    output, raise_type, exit_code = helpers.execute_error(command)

    assert raise_type == SystemExit
    assert exit_code == 2
    assert output[0:len(expect)] == expect


def test_generate_jsondata_without_args():
    """Check the json data generated is correct."""
    result = generate_jsondata(
        scope='foo',
        name='bar',
        identifier='1234',
        schema='234',
        args=argparse.Namespace(bin='', script=''),
    )

    assert result == {
        'description': '234 package created using A28 cmd by foo.',
        'identifier': '1234',
        'name': '@foo/bar',
        'version': '0.1.0',
    }


def test_generate_jsondata_with_bin():
    """Check the json data generated is correct with binary dir."""
    result = generate_jsondata(
        scope='foo',
        name='bar',
        identifier='1234',
        schema='234',
        args=argparse.Namespace(bin='in the bin', script=''),
    )

    assert result == {
        'description': '234 package created using A28 cmd by foo.',
        'identifier': '1234',
        'name': '@foo/bar',
        'version': '0.1.0',
        'bin': {},
    }


def test_generate_jsondata_with_bin_and_script():
    """Check the json data generated is correct with script and bin."""
    result = generate_jsondata(
        scope='foo',
        name='bar',
        identifier='1234',
        schema='234',
        args=argparse.Namespace(bin='in the bin', script='the script'),
    )

    assert result == {
        'description': '234 package created using A28 cmd by foo.',
        'identifier': '1234',
        'name': '@foo/bar',
        'version': '0.1.0',
        'bin': {},
        'scripts': {},
    }



"""Build packages to be submitted to Area28."""
import argparse
import sys

from a28 import __version__, account, api, package, system


def build_parser():
    """Build the argument for the cli."""
    parser = argparse.ArgumentParser(
        'a28',
        description='Area28 development kit'
    )
    app = '%(prog)s version ' + __version__
    parser.add_argument('-v', '--version', action='version', version=app)
    sub = parser.add_subparsers(dest='action', required=True, help='actions')

    api.cli_options(sub)
    package.cli_options(sub)
    system.cli_options(sub)
    account.cli_options(sub)

    return parser


def main(args=None):
    """Command line entry point."""
    args = args or sys.argv[1:]
    parser = build_parser()

    try:
        args = parser.parse_args(args)
        args.func(args)
        raise SystemExit(0)

    except Exception:
        sys.exit(1)



"""Build packages to be submitted to Area28."""
import logging

import coloredlogs

from a28.__version__ import __version__  # noqa: F401

# enable logging
logging.basicConfig()
log = logging.getLogger()
log.setLevel(logging.DEBUG)
coloredlogs.install(level='DEBUG', logger=log)



# vim: encoding=utf-8 ts=4 et sts=4 sw=4 tw=79 fileformat=unix nu wm=2
"""Test the a28 utilities."""
import pytest

from a28 import utils


def test_string_validation_short() -> None:
    """Test if a string is to short."""
    assert utils.valid_string('ab', min_length=3) is False


def test_string_validation_long() -> None:
    """Test if a string is to long."""
    assert utils.valid_string('abcdefghijklmnop', max_length=10) is False


def test_string_validation_invalid() -> None:
    """Test to ensure valid characters are not allowed."""
    assert utils.valid_string('!@£$%^&*(())__+') is False


def test_string_validation_valid() -> None:
    """Test a string with valid characters."""
    assert utils.valid_string('avalidword') is True


def test_confirm_nope(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test negative confirmation."""
    # monkeypatch the "input" function, so that it returns "Mark".
    # This simulates the user entering "Mark" in the terminal:
    monkeypatch.setattr('builtins.input', lambda _: 'nope')
    result = utils.confirm('test message')

    assert result is False


def test_confirm_y(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test positive confirmation."""
    # monkeypatch the "input" function, so that it returns "Mark".
    # This simulates the user entering "Mark" in the terminal:
    monkeypatch.setattr('builtins.input', lambda _: 'y')
    result = utils.confirm('test message')

    assert result is True


def test_confirm_yes(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test positive confirmation."""
    # monkeypatch the "input" function, so that it returns "Mark".
    # This simulates the user entering "Mark" in the terminal:
    monkeypatch.setattr('builtins.input', lambda _: 'yes')
    result = utils.confirm('test message')

    assert result is True


def test_confirm_ok(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test positive confirmation."""
    # monkeypatch the "input" function, so that it returns "Mark".
    # This simulates the user entering "Mark" in the terminal:
    monkeypatch.setattr('builtins.input', lambda _: 'ok')
    result = utils.confirm('test message')

    assert result is True


def test_confirm_sure(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test positive confirmation."""
    # monkeypatch the "input" function, so that it returns "Mark".
    # This simulates the user entering "Mark" in the terminal:
    monkeypatch.setattr('builtins.input', lambda _: 'sure')
    result = utils.confirm('test message')

    assert result is True


def test_confirm_alrighty(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test positive confirmation."""
    # monkeypatch the "input" function, so that it returns "Mark".
    # This simulates the user entering "Mark" in the terminal:
    monkeypatch.setattr('builtins.input', lambda _: 'alrighty')
    result = utils.confirm('test message')

    assert result is True


def test_confirm_ofcourse(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test positive confirmation."""
    # monkeypatch the "input" function, so that it returns "Mark".
    # This simulates the user entering "Mark" in the terminal:
    monkeypatch.setattr('builtins.input', lambda _: 'of course')
    result = utils.confirm('test message')

    assert result is True


def test_confirm_yup(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test positive confirmation."""
    # monkeypatch the "input" function, so that it returns "Mark".
    # This simulates the user entering "Mark" in the terminal:
    monkeypatch.setattr('builtins.input', lambda _: 'yup')
    result = utils.confirm('test message')

    assert result is True
